package types

import (
	"sync"
	"time"
)

type TTLMap[K comparable, V any] struct {
	rw  sync.RWMutex
	m   map[K]entry[V]
	ttl time.Duration
}

type entry[V any] struct {
	value    V
	expireAt *time.Time
}

func (e *entry[V]) expired() bool {
	if e.expireAt == nil {
		return false
	}
	return time.Now().After(*e.expireAt)
}

func NewTTLMap[K comparable, V any](ttl time.Duration) *TTLMap[K, V] {
	return &TTLMap[K, V]{m: make(map[K]entry[V]), ttl: ttl}
}

func (t *TTLMap[K, V]) Set(k K, v V) {
	t.SetWithTTL(k, v, t.ttl)
}

func (t *TTLMap[K, V]) SetWithTTL(k K, v V, ttl time.Duration) {
	defer t.rw.Unlock()
	t.rw.Lock()
	var expireAt *time.Time
	if ttl > 0 {
		at := time.Now().Add(ttl)
		expireAt = &at
	}
	t.m[k] = entry[V]{value: v, expireAt: expireAt}
}

func (t *TTLMap[K, V]) Get(k K) (val V, ok bool) {
	defer t.rw.RUnlock()
	t.rw.RLock()
	v, ok := t.m[k]
	if !ok {
		return
	}
	if v.expired() {
		defer func() {
			t.rw.RUnlock()
			t.rw.Lock()
			delete(t.m, k)
			t.rw.Unlock()
			t.rw.RLock()
		}()
		return
	}
	return v.value, true
}
